@@ -3,28 +3,28 @@ A low-level PDF generator.
 
 """
 
+import base64
 import re
 import zlib
 from codecs import BOM_UTF16_BE
 from hashlib import md5
 from math import ceil, log
+from warnings import warn
 
-VERSION = __version__ = '0.7.0'
+VERSION = __version__ = '0.11.0'
 
 
 def _to_bytes(item):
     """Convert item to bytes."""
     if isinstance(item, bytes):
         return item
-    elif isinstance(item, Object):
-        return item.data
     elif isinstance(item, float):
         if item.is_integer():
-            return f'{int(item):d}'.encode('ascii')
+            return str(int(item)).encode('ascii')
         else:
             return f'{item:f}'.rstrip('0').encode('ascii')
-    elif isinstance(item, int):
-        return f'{item:d}'.encode('ascii')
+    elif isinstance(item, Object):
+        return item.data
     return str(item).encode('ascii')
 
 
@@ -122,13 +122,9 @@ class Stream(Object):
         self.stream.append(b'h')
 
     def color_space(self, space, stroke=False):
-        """Set the nonstroking color space.
-
-        If stroke is set to ``True``, set the stroking color space instead.
-
-        """
-        self.stream.append(
-            b'/' + _to_bytes(space) + b' ' + (b'CS' if stroke else b'cs'))
+        """color_space is deprecated, use set_color_space instead."""
+        warn(Stream.color_space.__doc__, DeprecationWarning)
+        self.set_color_space(space, stroke)
 
     def curve_to(self, x1, y1, x2, y2, x3, y3):
         """Add cubic Bézier curve to current path.
@@ -204,6 +200,37 @@ class Stream(Object):
         """
         self.stream.append(b'b*' if even_odd else b'b')
 
+    def inline_image(self, width, height, color_space, bpc, raw_data):
+        """Add an inline image.
+
+        :param width: The width of the image.
+        :type width: :obj:`int`
+        :param height: The height of the image.
+        :type height: :obj:`int`
+        :param colorspace: The color space of the image, f.e. RGB, Gray.
+        :type colorspace: :obj:`str`
+        :param bpc: The bits per component. 1 for BW, 8 for grayscale.
+        :type bpc: :obj:`int`
+        :param raw_data: The raw pixel data.
+
+        """
+        data = zlib.compress(raw_data) if self.compress else raw_data
+        a85_data = base64.a85encode(data) + b'~>'
+        self.stream.append(b' '.join((
+            b'BI',
+            b'/W', _to_bytes(width),
+            b'/H', _to_bytes(height),
+            b'/BPC', _to_bytes(bpc),
+            b'/CS',
+            b'/Device' + _to_bytes(color_space),
+            b'/F',
+            b'[/A85 /Fl]' if self.compress else b'/A85',
+            b'/L', _to_bytes(len(a85_data)),
+            b'ID',
+            a85_data,
+            b'EI',
+        )))
+
     def line_to(self, x, y):
         """Add line from current point to point ``(x, y)``."""
         self.stream.append(b' '.join((_to_bytes(x), _to_bytes(y), b'l')))
@@ -217,6 +244,11 @@ class Stream(Object):
         self.stream.append(b' '.join((_to_bytes(x), _to_bytes(y), b'Td')))
 
     def shading(self, name):
+        """shading is deprecated, use paint_shading instead."""
+        warn(Stream.shading.__doc__, DeprecationWarning)
+        self.paint_shading(name)
+
+    def paint_shading(self, name):
         """Paint shape and color shading using shading dictionary ``name``."""
         self.stream.append(b'/' + _to_bytes(name) + b' sh')
 
@@ -250,14 +282,26 @@ class Stream(Object):
             _to_bytes(r), _to_bytes(g), _to_bytes(b),
             (b'RG' if stroke else b'rg'))))
 
-    def set_color_special(self, name, stroke=False):
-        """Set color for nonstroking operations.
+    def set_color_space(self, space, stroke=False):
+        """Set the nonstroking color space.
 
-        Set color for stroking operation if ``stroke`` is set to ``True``.
+        If stroke is set to ``True``, set the stroking color space instead.
 
         """
         self.stream.append(
-            b'/' + _to_bytes(name) + b' ' + (b'SCN' if stroke else b'scn'))
+            b'/' + _to_bytes(space) + b' ' + (b'CS' if stroke else b'cs'))
+
+    def set_color_special(self, name, stroke=False, *operands):
+        """Set special color for nonstroking operations.
+
+        Set special color for stroking operation if ``stroke`` is set to ``True``.
+
+        """
+        if name:
+            operands = (*operands, b'/' + _to_bytes(name))
+        self.stream.append(
+            b' '.join(_to_bytes(operand) for operand in operands) + b' ' +
+            (b'SCN' if stroke else b'scn'))
 
     def set_dash(self, dash_array, dash_phase):
         """Set dash line pattern.
@@ -280,6 +324,10 @@ class Stream(Object):
         """Set text rendering mode."""
         self.stream.append(_to_bytes(mode) + b' Tr')
 
+    def set_text_rise(self, height):
+        """Set text rise."""
+        self.stream.append(_to_bytes(height) + b' Ts')
+
     def set_line_cap(self, line_cap):
         """Set line cap style."""
         self.stream.append(_to_bytes(line_cap) + b' J')
@@ -292,6 +340,27 @@ class Stream(Object):
         """Set line width."""
         self.stream.append(_to_bytes(width) + b' w')
 
+    def set_matrix(self, a, b, c, d, e, f):
+        """Set current transformation matrix.
+
+        :param a: Top left number in the matrix.
+        :type a: :obj:`int` or :obj:`float`
+        :param b: Top middle number in the matrix.
+        :type b: :obj:`int` or :obj:`float`
+        :param c: Middle left number in the matrix.
+        :type c: :obj:`int` or :obj:`float`
+        :param d: Middle middle number in the matrix.
+        :type d: :obj:`int` or :obj:`float`
+        :param e: Bottom left number in the matrix.
+        :type e: :obj:`int` or :obj:`float`
+        :param f: Bottom middle number in the matrix.
+        :type f: :obj:`int` or :obj:`float`
+
+        """
+        self.stream.append(b' '.join((
+            _to_bytes(a), _to_bytes(b), _to_bytes(c),
+            _to_bytes(d), _to_bytes(e), _to_bytes(f), b'cm')))
+
     def set_miter_limit(self, miter_limit):
         """Set miter limit."""
         self.stream.append(_to_bytes(miter_limit) + b' M')
@@ -304,24 +373,8 @@ class Stream(Object):
         """
         self.stream.append(b'/' + _to_bytes(state_name) + b' gs')
 
-    def show_text(self, text):
-        """Show text strings with individual glyph positioning."""
-        self.stream.append(b'[' + _to_bytes(text) + b'] TJ')
-
-    def show_text_string(self, text):
-        """Show single text string."""
-        self.stream.append(String(text).data + b' Tj')
-
-    def stroke(self):
-        """Stroke path."""
-        self.stream.append(b'S')
-
-    def stroke_and_close(self):
-        """Stroke and close path."""
-        self.stream.append(b's')
-
-    def text_matrix(self, a, b, c, d, e, f):
-        """Set text matrix and text line matrix.
+    def set_text_matrix(self, a, b, c, d, e, f):
+        """Set current text and text line transformation matrix.
 
         :param a: Top left number in the matrix.
         :type a: :obj:`int` or :obj:`float`
@@ -341,26 +394,31 @@ class Stream(Object):
             _to_bytes(a), _to_bytes(b), _to_bytes(c),
             _to_bytes(d), _to_bytes(e), _to_bytes(f), b'Tm')))
 
+    def text_matrix(self, a, b, c, d, e, f):
+        """text_matrix is deprecated, use set_text_matrix instead."""
+        warn(Stream.text_matrix.__doc__, DeprecationWarning)
+        self.set_text_matrix(a, b, c, d, e, f)
+
     def transform(self, a, b, c, d, e, f):
-        """Modify current transformation matrix.
+        """transform is deprecated, use set_matrix instead."""
+        warn(Stream.transform.__doc__, DeprecationWarning)
+        self.set_matrix(a, b, c, d, e, f)
 
-        :param a: Top left number in the matrix.
-        :type a: :obj:`int` or :obj:`float`
-        :param b: Top middle number in the matrix.
-        :type b: :obj:`int` or :obj:`float`
-        :param c: Middle left number in the matrix.
-        :type c: :obj:`int` or :obj:`float`
-        :param d: Middle middle number in the matrix.
-        :type d: :obj:`int` or :obj:`float`
-        :param e: Bottom left number in the matrix.
-        :type e: :obj:`int` or :obj:`float`
-        :param f: Bottom middle number in the matrix.
-        :type f: :obj:`int` or :obj:`float`
+    def show_text(self, text):
+        """Show text strings with individual glyph positioning."""
+        self.stream.append(b'[' + _to_bytes(text) + b'] TJ')
 
-        """
-        self.stream.append(b' '.join((
-            _to_bytes(a), _to_bytes(b), _to_bytes(c),
-            _to_bytes(d), _to_bytes(e), _to_bytes(f), b'cm')))
+    def show_text_string(self, text):
+        """Show single text string."""
+        self.stream.append(String(text).data + b' Tj')
+
+    def stroke(self):
+        """Stroke path."""
+        self.stream.append(b'S')
+
+    def stroke_and_close(self):
+        """Stroke and close path."""
+        self.stream.append(b's')
 
     @property
     def data(self):
@@ -409,17 +467,8 @@ class Array(Object, list):
 
 class PDF:
     """PDF document."""
-    def __init__(self, version=b'1.7', identifier=None):
-        """Create a PDF document.
-
-        :param bytes version: PDF version.
-        :param bytes identifier: PDF file identifier.
-
-        """
-        #: PDF version, as :obj:`bytes`.
-        self.version = _to_bytes(version)
-        #: PDF file identifier.
-        self.identifier = identifier
+    def __init__(self):
+        """Create a PDF document."""
 
         #: Python :obj:`list` containing the PDF’s objects.
         self.objects = []
@@ -487,18 +536,23 @@ class PDF:
         self.current_position += len(content) + 1
         output.write(content + b'\n')
 
-    def write(self, output, version=None, identifier=None, compress=False):
+    def write(self, output, version=b'1.7', identifier=False, compress=False):
         """Write PDF to output.
 
         :param output: Output stream.
         :type output: binary :term:`file object`
         :param bytes version: PDF version.
-        :param bytes identifier: PDF file identifier.
+        :param identifier: PDF file identifier. Default is :obj:`False`
+          to include no identifier, can be set to :obj:`True` to generate an
+          automatic identifier.
+        :type identifier: :obj:`bytes` or :obj:`bool`
         :param bool compress: whether the PDF uses a compressed object stream.
 
         """
-        version = self.version if version is None else _to_bytes(version)
-        identifier = self.identifier if identifier is None else identifier
+        # Convert version and identifier to bytes
+        version = _to_bytes(version or b'1.7')  # Force 1.7 when None
+        if identifier not in (False, True, None):
+            identifier = _to_bytes(identifier)
 
         # Write header
         self.write_line(b'%PDF-' + version, output)
@@ -566,10 +620,12 @@ class PDF:
                 'Root': self.catalog.reference,
                 'Info': self.info.reference,
             }
-            if identifier is not None:
+            if identifier:
                 data = b''.join(
                     obj.data for obj in self.objects if obj.free != 'f')
                 data_hash = md5(data).hexdigest().encode()
+                if identifier is True:
+                    identifier = data_hash
                 extra['ID'] = Array((
                     String(identifier).data, String(data_hash).data))
             dict_stream = Stream([xref_stream], extra, compress)
@@ -599,10 +655,12 @@ class PDF:
             self.write_line(f'/Size {len(self.objects)}'.encode(), output)
             self.write_line(b'/Root ' + self.catalog.reference, output)
             self.write_line(b'/Info ' + self.info.reference, output)
-            if identifier is not None:
+            if identifier:
                 data = b''.join(
                     obj.data for obj in self.objects if obj.free != 'f')
                 data_hash = md5(data).hexdigest().encode()
+                if identifier is True:
+                    identifier = data_hash
                 self.write_line(
                     b'/ID [' + String(identifier).data + b' ' +
                     String(data_hash).data + b']', output)
